from unittest import TestCase
from iotio.PacketEncoder import PacketDataType, DefaultPacketEncoder


class TestPacketDataType(TestCase):
    def test_type_to_byte(self):
        self.assertEqual(PacketDataType.type_to_byte(b'\xFF'), PacketDataType.BINARY.to_bytes(1, "big"))
        self.assertEqual(PacketDataType.type_to_byte(True), PacketDataType.BOOLEAN.to_bytes(1, "big"))
        self.assertEqual(PacketDataType.type_to_byte(20), PacketDataType.INTEGER.to_bytes(1, "big"))
        self.assertEqual(PacketDataType.type_to_byte("data"), PacketDataType.STRING.to_bytes(1, "big"))
        self.assertEqual(PacketDataType.type_to_byte({"a": 5}), PacketDataType.JSON.to_bytes(1, "big"))


class TestDefaultPacketEncoder(TestCase):
    binary = b'\x45\x0F\x24'
    binary_enc = None

    boolean = True
    boolean_enc = None

    integer_1 = 15409255571
    integer_1_enc = None

    integer_2 = -1204519556
    integer_2_enc = None

    string = "some_data"
    string_enc = None

    json = {
        "test": 5,
        "a": ["a", 2, False, {"a": None}],
        "b": None,
        "c": True
    }
    json_enc = None

    def test_encode(self):
        try:
            self.binary_enc = DefaultPacketEncoder.encode("test", self.binary)
            self.boolean_enc = DefaultPacketEncoder.encode("test", self.boolean)
            self.integer_1_enc = DefaultPacketEncoder.encode("test", self.integer_1)
            self.integer_2_enc = DefaultPacketEncoder.encode("test", self.integer_2)
            self.string_enc = DefaultPacketEncoder.encode("test", self.string)
            self.json_enc = DefaultPacketEncoder.encode("test", self.json)
        except ValueError:
            self.fail()

    def test_decode(self):
        # ensure encoding is completed for test decoding is done
        self.test_encode()

        self.assertEqual(self.binary, DefaultPacketEncoder.decode(self.binary_enc)[1])
        self.assertEqual(self.boolean, DefaultPacketEncoder.decode(self.boolean_enc)[1])
        self.assertEqual(self.integer_1, DefaultPacketEncoder.decode(self.integer_1_enc)[1])
        self.assertEqual(self.integer_2, DefaultPacketEncoder.decode(self.integer_2_enc)[1])
        self.assertEqual(self.string, DefaultPacketEncoder.decode(self.string_enc)[1])
        self.assertEqual(self.json, DefaultPacketEncoder.decode(self.json_enc)[1])



# default
from typing import TYPE_CHECKING, Tuple, Union
if TYPE_CHECKING:
    from .Manager import IoTManager
import logging

# external
from eventlet.websocket import WebSocket

# internal
from .PacketEncoder import AbstractPacketEncoder, DefaultPacketEncoder
from .types import sendable
from .Endpoint import EndpointManager, EndpointParseResponse, ValidationResponse, AbstractEndpointValidator


# iot client class
class IoTClient:
    def __init__(self, ws: WebSocket, client_id: str, client_type: str, client_data: dict, manager: 'IoTManager',
                 logging_level: int = logging.ERROR, encoder: AbstractPacketEncoder = DefaultPacketEncoder):
        """
        Class which represents an active connection by a IoTClient and is used to manage that connection.

        :param ws: WebSocket object which is the connection used to communicate between the client and server.
        :param client_id: ID of the connecting client, should be a unique identifier which can be used to identify the
                          client over multiple connections.
        :param client_type: The type of the client, used to separate different types of connecting devices.
        :param client_data: JSON data provided by the client at time of connection.
        :param logging_level: Logging level of the instance, useful for debugging.
        :param encoder: A instance of a AbstractPacketEncoder implementation, used to define how packets should be
                        encoded/decoded from binary by the server when communicating with the client. Shouldn't be
                        changed unless the user has a good understanding of how the iot.io protocol works.
        """
        self.logger = logging.Logger("[iot.io.client:" + client_id + "]")
        self.logger.level = logging_level

        # reference to the client's websocket connection
        self.socket = ws

        # reference to the IoTManager
        self.manager = manager

        # encoder used for encoding and decoding packets
        self.encoder = encoder

        # the list of rooms the client is in
        self.__rooms = []

        # save the id and type
        self.__id = client_id
        self.__type = client_type
        self.__data = client_data

        # endpoint validation manager
        self.__endpoint_manager = EndpointManager()

    # properties for id, type, and data
    @property
    def id(self):
        return self.__id

    @property
    def type(self):
        return self.__type

    @property
    def data(self):
        return self.__data

    @property
    def endpoints(self):
        return self.__endpoint_manager.endpoints

    @property
    def rooms(self):
        return self.__rooms

    @property
    def info(self):
        return {
            "id": self.id,
            "type": self.type,
            "data": self.data,
            "endpoints": self.endpoints
        }

    # get the endpoint_validator
    def get_validator(self, endpoint_id: str) -> Union[AbstractEndpointValidator, ValidationResponse]:
        if self.__endpoint_manager.initialized:
            return self.__endpoint_manager.get_validator(endpoint_id=endpoint_id)
        else:
            return ValidationResponse.CLIENT_ENDPOINTS_NOT_INITIALIZED

    def parse_endpoints(self, endpoints) -> Tuple[str, EndpointParseResponse]:
        return self.__endpoint_manager.parse(endpoints)

    def emit(self, event: str, data: sendable, trigger_update: bool = True):
        """
        Emit function for sending data to the client.

        :param event: The client side event to send the data to.
        :param data: The data to be sent to the client(s).
        :param trigger_update: If this emit should trigger the managers on_update handler. This is useful for preventing
                               unnecessary triggers of the on_update function if for example it is used to update a
                               database for data persistence.
        :return:
        """
        if not isinstance(event, str):
            raise TypeError("'event' must be of type str")

        if not self.socket.websocket_closed:
            self.logger.debug("Sending a message for event '", event, "' with type: '", type(data), "'")

            self.socket.send(self.encoder.encode(event, data))

            if trigger_update:
                self.manager.log_update(self, event, data)

    def join(self, room: str, **kwargs):
        """
        Have the client join the specified room.

        :param room: ID of the room to be joined as a string
        :return:
        """
        if not kwargs.pop("called_by_manager", False):
            self.manager.join(self, room, called_by_client=True)

        self.__rooms.append(room)

    def leave(self, room: str, **kwargs):
        """
        Have the client leave the specified room.

        :param room: ID of the room to be left as a string
        :return:
        """
        if not kwargs.pop("called_by_manager", False):
            self.manager.leave(self, room, called_by_client=True)

        self.__rooms.remove(room)



import enum


# enumerator for errors between client and server
class Errors(enum.Enum):
    CLIENT_NO_ID = "client_no_id"
    CLIENT_NO_TYPE = "client_no_type"
    CLIENT_INVALID_TYPE = "client_invalid_type"
    CLIENT_INVALID_DATA = "client_invalid_data"
    CLIENT_INVALID_ENDPOINTS = "client_invalid_endpoints"
    CLIENT_NO_PROTOCOL_VERSION = "client_no_protocol_version"
    CLIENT_INCOMPATIBLE_PROTOCOL_VERSION = "client_incompatible_protocol_version"
    CLIENT_INVALID_PACKET = "client_invalid_packet"



# default
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .Manager import IoTManager

# internal
from .Client import IoTClient
from .types import event_pair


class DeviceType:
    def __init__(self, type_name: str):
        """
        Type of device as a string, should match the device_type string provided by clients when they connect.

        :param type_name: Device type as a string.
        """
        # the type of device
        self.__type = type_name

        # a reference to the manager the type is running under
        self.__context = None

    @property
    def type(self):
        return self.__type

    @property
    def context(self) -> 'IoTManager':
        return self.__context

    def set_context(self, context: 'IoTManager'):
        """
        Used to set the context of the device type. Used by the IoTManager.

        :return:
        """
        self.__context = context

    def call_event_handler(self, event: str, message: str, client: IoTClient) -> event_pair:
        """
        Calls the event handler for the given event if it exists, if it doesnt exist then it does nothing.

        Don't overwrite.

        :param event: The event which is being invoked.
        :param message: The message to pass to the event handler.
        :param client: The client which sent the message.
        :return: None or str
        """
        try:
            response = getattr(self, "on_" + event)(message, client)

            if response is None:
                return None, None
            else:
                if isinstance(response, tuple):
                    return response[0], response[1]
                else:
                    return event, response
        except AttributeError:
            return None, None

    def on_connect(self, client: IoTClient):
        pass

    def on_disconnect(self, client: IoTClient):
        pass



# iot error base
class IoTError(Exception):
    """
    Base Exception for all iot.io errors
    """
    def __init__(self, message: str = "Base exception class for the iot.io library."):
        super().__init__(message)


# error when a client has its connection ended
class ConnectionEnded(IoTError):
    """
    Exception thrown when an operation attempts to use a socket but the connection is no longer alive.
    """
    def __init__(self, message: str = "Client socket no longer alive."):
        super().__init__(message)


# generic connection failed error for when client does something wrong while first connecting
class ConnectionFailed(IoTError):
    """
    Exception thrown when a client does something wrong while connecting, this is usually caused by missing or
    invalid HTTP Headers. Should not normally be seen unless you are modifying the IoTManager or IoTManagerMiddleware
    classes directly.
    """
    def __init__(self, message: str = "Client failed to connect to the server."):
        super().__init__(message)


# error when client does not provide an Id when connecting
class ClientNoId(ConnectionFailed):
    """
    Exception thrown when a connecting client does not provide an ID.
    """
    def __init__(self, message: str = "Client failed to provide the IoT-IO-Id header field."):
        super().__init__(message)


# error when client does not provide a Type when connecting
class ClientNoType(ConnectionFailed):
    """
    Exception thrown when a connecting client does not provide a Type.
    """
    def __init__(self, message: str = "Client failed to provide the IoT-IO-Type header field."):
        super().__init__(message)


# error when client provides an invalid Type when connecting
class ClientInvalidType(ConnectionFailed):
    """
    Exception thrown when a connecting client does not provide a Type which the server can handle.
    """
    def __init__(self, message: str = "Client provided an invalid value for the IoT-IO-Type header field."):
        super().__init__(message)


# error when client provides invalid json in the IoT-IO-Data field
class ClientInvalidData(ConnectionFailed):
    """
    Exception thrown when a connecting client does not provide Data which the server can deserialize.
    """
    def __init__(self, message: str = "Client provided an invalid json data for the IoT-IO-Data header field."):
        super().__init__(message)


# error when client provides invalid data in the IoT-IO-Endpoints field
class ClientInvalidEndpoints(ConnectionFailed):
    """
    Exception thrown when a connecting client does not provide Endpoints which the server can understand.
    """
    def __init__(self, message: str = "Client provided an invalid value for the IoT-IO-Endpoints header field."):
        super().__init__(message)


# error when client does not provide a protocol version
class ClientNoProtocolVersion(ConnectionFailed):
    """
    Exception thrown when a connecting client does not provide it's protocol version.
    """

    def __init__(self, message: str = "Client did not provide the IoT-IO-ProtocolVersion header field."):
        super().__init__(message)


# error when client provides invalid data in the IoT-IO-Endpoints field
class ClientIncompatibleProtocolVersion(ConnectionFailed):
    """
    Exception thrown when a connecting client's protocol version is incompatible with that of the server's.
    """

    def __init__(self, message: str = "The version of the client's protocol is incompatible with that of the server."):
        super().__init__(message)



try:
    # default
    import functools
    from typing import Callable, TYPE_CHECKING

    if TYPE_CHECKING:
        from .Manager import IoTManager

    # external
    import flask
    from flask_restful import Resource, reqparse

    # internal
    from .Endpoint import ValidationResponse, EndpointType


    class EndpointResource(Resource):
        def __init__(self, manager: 'IoTManager', auth_decorator: Callable):
            self.manager = manager

            # decorate post method with the provided decorator
            if auth_decorator:
                self.post = auth_decorator(self.post)

            self.post_parser = reqparse.RequestParser()
            self.post_parser.add_argument("clientId", location=["json", "args"], required=True, type=str,
                                          help="ID of Client.")
            self.post_parser.add_argument("endpointId", location=["json", "args"], required=True, type=str,
                                          help="ID of Endpoint where data should be sent.")
            self.post_parser.add_argument("data", location=["json", "args"], required=True, help="Data to be sent.")

        def post(self):
            args = self.post_parser.parse_args()

            # get validator for the specified client and endpoint
            validator = self.manager.get_validator(args["clientId"], args["endpointId"])

            # check if validator was found
            if isinstance(validator, ValidationResponse):
                return {
                    "success": False,
                    "error": validator.value
                }, 400

            # try and validate the data
            response, data = validator.validate(args["data"])

            # if validation failed
            if response != ValidationResponse.VALID:
                return {
                    "success": False,
                    "error": response.value
                }, 400

            # if validation worked then send the data and return a success
            self.manager.emit(validator.id, data, client_id=args["clientId"])

            return {
                "success": True
            }

except ImportError:
    EndpointResource = None



from unittest import TestCase
from iotio.Endpoint import ValidationResponse, EndpointType, AbstractEndpointValidator, BooleanEndpointValidator, \
    IntegerEndpointValidator, StringEndpointValidator, EnumEndpointValidator, EndpointParseResponse


class TestAbstractEndpointValidator(TestCase):
    validator = AbstractEndpointValidator("abc", "def", "ghi")

    def test_validate(self):
        self.assertIsNone(self.validator.validate(None))

    def test_parse(self):
        val_1 = self.validator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c"
        })
        val_1: AbstractEndpointValidator

        self.assertIsNotNone(val_1)
        self.assertEqual(val_1.id, "a")
        self.assertEqual(val_1.name, "b")
        self.assertEqual(val_1.special_id, "c")

        val_2 = self.validator.parse({
            "id": "d",
            "name": "e",
            "specialId": None
        })
        val_2: AbstractEndpointValidator

        self.assertIsNotNone(val_2)
        self.assertEqual(val_2.id, "d")
        self.assertEqual(val_2.name, "e")
        self.assertEqual(val_2.special_id, None)

        self.assertEqual(self.validator.parse({
            "id": 1,
            "name": "b",
            "specialId": "c"
        }), EndpointParseResponse.ID_MUST_BE_STRING)

        self.assertEqual(self.validator.parse({
            "id": "a",
            "name": 1,
            "specialId": "c"
        }), EndpointParseResponse.NAME_MUST_BE_STRING)

        self.assertEqual(self.validator.parse({
            "id": "",
            "name": "b",
            "specialId": 1
        }), EndpointParseResponse.SPECIAL_ID_MUST_BE_STRING)

    def test__dict__(self):
        data = self.validator.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("specialId", None), "ghi")


class TestBooleanEndpointValidator(TestCase):
    validator = BooleanEndpointValidator("abc", "def", "ghi")

    def test_validate(self):
        self.assertEqual(self.validator.validate(True), ValidationResponse.VALID)
        self.assertEqual(self.validator.validate("test"), ValidationResponse.INVALID_TYPE)

    def test_parse(self):
        val_1 = self.validator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c"
        })

        self.assertIsNotNone(val_1)
        self.assertEqual(val_1.id, "a")
        self.assertEqual(val_1.name, "b")
        self.assertEqual(val_1.special_id, "c")

        val_2 = self.validator.parse({
            "id": "d",
            "name": "e",
            "specialId": None
        })

        self.assertIsNotNone(val_2)
        self.assertEqual(val_2.id, "d")
        self.assertEqual(val_2.name, "e")
        self.assertIsNone(val_2.special_id)

        self.assertEqual(self.validator.parse({
            "id": 1,
            "name": "b",
            "specialId": "c"
        }), EndpointParseResponse.ID_MUST_BE_STRING)

        self.assertEqual(self.validator.parse({
            "id": "a",
            "name": 1,
            "specialId": "c"
        }), EndpointParseResponse.NAME_MUST_BE_STRING)

        self.assertEqual(self.validator.parse({
            "id": "",
            "name": "b",
            "specialId": 1
        }), EndpointParseResponse.SPECIAL_ID_MUST_BE_STRING)

    def test__dict__(self):
        data = self.validator.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.BOOLEAN.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNone(data.get("constraints", ""))


class TestNumberEndpointValidator(TestCase):
    validator_1 = IntegerEndpointValidator("abc", "def", "ghi", 5, 10)
    validator_2 = IntegerEndpointValidator("abc", "def", "ghi", 5)
    validator_3 = IntegerEndpointValidator("abc", "def", "ghi", None, 10)
    validator_4 = IntegerEndpointValidator("abc", "def", "ghi", 5, 10, 1)
    validator_5 = IntegerEndpointValidator("abc", "def", "ghi")

    def test_validate(self):
        self.assertEqual(self.validator_1.validate(7), ValidationResponse.VALID)
        self.assertEqual(self.validator_1.validate(3), ValidationResponse.LOWER_THAN_MIN)
        self.assertEqual(self.validator_1.validate(12), ValidationResponse.HIGHER_THAN_MAX)
        self.assertEqual(self.validator_1.validate(""), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_2.validate(7), ValidationResponse.VALID)
        self.assertEqual(self.validator_2.validate(3), ValidationResponse.LOWER_THAN_MIN)
        self.assertEqual(self.validator_2.validate(12), ValidationResponse.VALID)
        self.assertEqual(self.validator_2.validate(""), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_3.validate(7), ValidationResponse.VALID)
        self.assertEqual(self.validator_3.validate(3), ValidationResponse.VALID)
        self.assertEqual(self.validator_3.validate(12), ValidationResponse.HIGHER_THAN_MAX)
        self.assertEqual(self.validator_3.validate(""), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_4.validate(7), ValidationResponse.VALID)
        self.assertEqual(self.validator_4.validate(3), ValidationResponse.LOWER_THAN_MIN)
        self.assertEqual(self.validator_4.validate(12), ValidationResponse.HIGHER_THAN_MAX)
        self.assertEqual(self.validator_4.validate(""), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_5.validate(7), ValidationResponse.VALID)
        self.assertEqual(self.validator_5.validate(3), ValidationResponse.VALID)
        self.assertEqual(self.validator_5.validate(12), ValidationResponse.VALID)
        self.assertEqual(self.validator_5.validate(""), ValidationResponse.INVALID_TYPE)

    def test_parse(self):
        val_1 = self.validator_1.parse(self.validator_1.__dict__())
        val_1: IntegerEndpointValidator

        self.assertEqual(val_1.id, self.validator_1.id)
        self.assertEqual(val_1.name, self.validator_1.name)
        self.assertEqual(val_1.special_id, self.validator_1.special_id)
        self.assertEqual(val_1.min, self.validator_1.min)
        self.assertEqual(val_1.max, self.validator_1.max)
        self.assertEqual(val_1.increment, self.validator_1.increment)

        val_2 = self.validator_2.parse(self.validator_2.__dict__())
        val_2: IntegerEndpointValidator

        self.assertEqual(val_2.id, self.validator_2.id)
        self.assertEqual(val_2.name, self.validator_2.name)
        self.assertEqual(val_2.special_id, self.validator_2.special_id)
        self.assertEqual(val_2.min, self.validator_2.min)
        self.assertEqual(val_2.max, self.validator_2.max)
        self.assertEqual(val_2.increment, self.validator_2.increment)

        val_3 = self.validator_3.parse(self.validator_3.__dict__())
        val_3: IntegerEndpointValidator

        self.assertEqual(val_3.id, self.validator_3.id)
        self.assertEqual(val_3.name, self.validator_3.name)
        self.assertEqual(val_3.special_id, self.validator_3.special_id)
        self.assertEqual(val_3.min, self.validator_3.min)
        self.assertEqual(val_3.max, self.validator_3.max)
        self.assertEqual(val_3.increment, self.validator_3.increment)

        val_4 = self.validator_4.parse(self.validator_4.__dict__())
        val_4: IntegerEndpointValidator

        self.assertEqual(val_4.id, self.validator_4.id)
        self.assertEqual(val_4.name, self.validator_4.name)
        self.assertEqual(val_4.special_id, self.validator_4.special_id)
        self.assertEqual(val_4.min, self.validator_4.min)
        self.assertEqual(val_4.max, self.validator_4.max)
        self.assertEqual(val_4.increment, self.validator_4.increment)

        val_5 = self.validator_5.parse(self.validator_5.__dict__())
        val_5: IntegerEndpointValidator

        self.assertEqual(val_5.id, self.validator_5.id)
        self.assertEqual(val_5.name, self.validator_5.name)
        self.assertEqual(val_5.special_id, self.validator_5.special_id)
        self.assertEqual(val_5.min, self.validator_5.min)
        self.assertEqual(val_5.max, self.validator_5.max)
        self.assertEqual(val_5.increment, self.validator_5.increment)

        self.assertEqual(IntegerEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "min": "",
                "max": 10,
                "increment": None
            }
        }), EndpointParseResponse.CONSTRAINT_MIN_MUST_BE_INT)

        self.assertEqual(IntegerEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "min": 5,
                "max": "",
                "increment": None
            }
        }), EndpointParseResponse.CONSTRAINT_MAX_MUST_BE_INT)

        self.assertEqual(IntegerEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "min": 5,
                "max": 10,
                "increment": ""
            }
        }), EndpointParseResponse.CONSTRAINT_INCREMENT_MUST_BE_INT)

    def test__dict__(self):
        data = self.validator_1.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.INTEGER.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("min", None), 5)
        self.assertEqual(data.get("constraints").get("max", None), 10)
        self.assertIsNone(data.get("constraints").get("increment", ""))

        data = self.validator_2.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.INTEGER.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("min", None), 5)
        self.assertIsNone(data.get("constraints").get("max", ""))
        self.assertIsNone(data.get("constraints").get("increment", None))

        data = self.validator_3.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.INTEGER.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertIsNone(data.get("constraints").get("min", ""))
        self.assertEqual(data.get("constraints").get("max", None), 10)
        self.assertIsNone(data.get("constraints").get("increment", ""))

        data = self.validator_4.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.INTEGER.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("min", None), 5)
        self.assertEqual(data.get("constraints").get("max", None), 10)
        self.assertEqual(data.get("constraints").get("increment", None), 1)

        data = self.validator_5.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.INTEGER.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertIsNone(data.get("constraints").get("min", ""))
        self.assertIsNone(data.get("constraints").get("max", ""))
        self.assertIsNone(data.get("constraints").get("increment", ""))


class TestStringEndpointValidator(TestCase):
    validator_1 = StringEndpointValidator("abc", "def", "ghi", 3, "abc")
    validator_2 = StringEndpointValidator("abc", "def", "ghi", 3, None, "abc")
    validator_3 = StringEndpointValidator("abc", "def", "ghi", 3, ["one", "two", "thr"])
    validator_4 = StringEndpointValidator("abc", "def", "ghi", 3, None, ["one", "two", "thr"])
    validator_5 = StringEndpointValidator("abc", "def", "ghi")

    def test_validate(self):
        self.assertEqual(self.validator_1.validate("abc"), ValidationResponse.VALID)
        self.assertEqual(self.validator_1.validate("a"), ValidationResponse.INVALID_LENGTH)
        self.assertEqual(self.validator_1.validate("def"), ValidationResponse.VIOLATES_ALLOW_LIST)
        self.assertEqual(self.validator_1.validate(1), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_2.validate("def"), ValidationResponse.VALID)
        self.assertEqual(self.validator_2.validate("a"), ValidationResponse.INVALID_LENGTH)
        self.assertEqual(self.validator_2.validate("abc"), ValidationResponse.VIOLATES_DENY_LIST)
        self.assertEqual(self.validator_2.validate(1), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_3.validate("one"), ValidationResponse.VALID)
        self.assertEqual(self.validator_3.validate("two"), ValidationResponse.VALID)
        self.assertEqual(self.validator_3.validate("thr"), ValidationResponse.VALID)
        self.assertEqual(self.validator_3.validate("one_two"), ValidationResponse.INVALID_LENGTH)
        self.assertEqual(self.validator_3.validate("foo"), ValidationResponse.VIOLATES_ALLOW_LIST)
        self.assertEqual(self.validator_3.validate(1), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_4.validate("foo"), ValidationResponse.VALID)
        self.assertEqual(self.validator_4.validate("one_two"), ValidationResponse.INVALID_LENGTH)
        self.assertEqual(self.validator_4.validate("one"), ValidationResponse.VIOLATES_DENY_LIST)
        self.assertEqual(self.validator_4.validate("two"), ValidationResponse.VIOLATES_DENY_LIST)
        self.assertEqual(self.validator_4.validate("thr"), ValidationResponse.VIOLATES_DENY_LIST)
        self.assertEqual(self.validator_4.validate(1), ValidationResponse.INVALID_TYPE)

        self.assertEqual(self.validator_5.validate("abc_def_ghi_jkl_mno"), ValidationResponse.VALID)
        self.assertEqual(self.validator_5.validate("qwe_rty_uio"), ValidationResponse.VALID)
        self.assertEqual(self.validator_5.validate(1), ValidationResponse.INVALID_TYPE)

    def test_parse(self):
        val_1 = self.validator_1.parse(self.validator_1.__dict__())
        val_1: StringEndpointValidator

        self.assertEqual(val_1.id, self.validator_1.id)
        self.assertEqual(val_1.name, self.validator_1.name)
        self.assertEqual(val_1.special_id, self.validator_1.special_id)
        self.assertEqual(val_1.length, self.validator_1.length)
        self.assertEqual(val_1.allow_list, self.validator_1.allow_list)
        self.assertEqual(val_1.deny_list, self.validator_1.deny_list)

        val_2 = self.validator_2.parse(self.validator_2.__dict__())
        val_2: StringEndpointValidator

        self.assertEqual(val_2.id, self.validator_2.id)
        self.assertEqual(val_2.name, self.validator_2.name)
        self.assertEqual(val_2.special_id, self.validator_2.special_id)
        self.assertEqual(val_2.length, self.validator_2.length)
        self.assertEqual(val_2.allow_list, self.validator_2.allow_list)
        self.assertEqual(val_2.deny_list, self.validator_2.deny_list)

        val_3 = self.validator_3.parse(self.validator_3.__dict__())
        val_3: StringEndpointValidator

        self.assertEqual(val_3.id, self.validator_3.id)
        self.assertEqual(val_3.name, self.validator_3.name)
        self.assertEqual(val_3.special_id, self.validator_3.special_id)
        self.assertEqual(val_3.length, self.validator_3.length)
        self.assertEqual(val_3.allow_list, self.validator_3.allow_list)
        self.assertEqual(val_3.deny_list, self.validator_3.deny_list)

        val_4 = self.validator_4.parse(self.validator_4.__dict__())
        val_4: StringEndpointValidator

        self.assertEqual(val_4.id, self.validator_4.id)
        self.assertEqual(val_4.name, self.validator_4.name)
        self.assertEqual(val_4.special_id, self.validator_4.special_id)
        self.assertEqual(val_4.length, self.validator_4.length)
        self.assertEqual(val_4.allow_list, self.validator_4.allow_list)
        self.assertEqual(val_4.deny_list, self.validator_4.deny_list)

        val_5 = self.validator_5.parse(self.validator_5.__dict__())
        val_5: StringEndpointValidator

        self.assertEqual(val_5.id, self.validator_5.id)
        self.assertEqual(val_5.name, self.validator_5.name)
        self.assertEqual(val_5.special_id, self.validator_5.special_id)
        self.assertEqual(val_5.length, self.validator_5.length)
        self.assertEqual(val_5.allow_list, self.validator_5.allow_list)
        self.assertEqual(val_5.deny_list, self.validator_5.deny_list)

        self.assertEqual(StringEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "length": "",
                "allowList": None,
                "denyList": None
            }
        }), EndpointParseResponse.CONSTRAINT_LENGTH_MUST_BE_INT)

        self.assertEqual(StringEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "length": 5,
                "allowList": "abc",
                "denyList": "edf"
            }
        }), EndpointParseResponse.CONSTRAINT_BOTH_ALLOW_LIST_AND_DENY_LIST_SPECIFIED)

        self.assertEqual(StringEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "length": 5,
                "allowList": 2,
                "denyList": None
            }
        }), EndpointParseResponse.CONSTRAINT_LIST_MUST_BE_STRING_OR_LIST)

        self.assertEqual(StringEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "length": 5,
                "allowList": None,
                "denyList": 5
            }
        }), EndpointParseResponse.CONSTRAINT_LIST_MUST_BE_STRING_OR_LIST)

        self.assertEqual(StringEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "length": 5,
                "allowList": ["a", 4],
                "denyList": None
            }
        }), EndpointParseResponse.CONSTRAINT_LIST_MUST_ONLY_CONTAIN_STRINGS)

        self.assertEqual(StringEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "length": 5,
                "allowList": None,
                "denyList": ["a", 5]
            }
        }), EndpointParseResponse.CONSTRAINT_LIST_MUST_ONLY_CONTAIN_STRINGS)

    def test__dict__(self):
        data = self.validator_1.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.STRING.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("length", None), 3)
        self.assertEqual(data.get("constraints").get("allowList", None), "abc")
        self.assertIsNone(data.get("constraints").get("denyList", ""))

        data = self.validator_2.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.STRING.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("length", None), 3)
        self.assertIsNone(data.get("constraints").get("allowList", ""))
        self.assertEqual(data.get("constraints").get("denyList", None), "abc")

        data = self.validator_3.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.STRING.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("length", None), 3)
        self.assertEqual(data.get("constraints").get("allowList", None), ["one", "two", "thr"])
        self.assertIsNone(data.get("constraints").get("denyList", ""))

        data = self.validator_4.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.STRING.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertEqual(data.get("constraints").get("length", None), 3)
        self.assertIsNone(data.get("constraints").get("allowList", ""))
        self.assertEqual(data.get("constraints").get("denyList", None), ["one", "two", "thr"])

        data = self.validator_5.__dict__()

        self.assertEqual(data.get("id", None), "abc")
        self.assertEqual(data.get("name", None), "def")
        self.assertEqual(data.get("type", None), EndpointType.STRING.value)
        self.assertEqual(data.get("specialId", None), "ghi")
        self.assertIsNotNone(data.get("constraints", None))
        self.assertIsNone(data.get("constraints").get("length", ""))
        self.assertIsNone(data.get("constraints").get("allowList", ""))
        self.assertIsNone(data.get("constraints").get("denyList", ""))


class TestEnumEndpointValidator(TestCase):
    validator_1 = EnumEndpointValidator("abc", "def", "ghi", {
        "Apple": "a",
        "Banana": "b",
        "Cucumber": "c"
    })

    validator_2 = EnumEndpointValidator("abc", "def", "ghi", {
        "Apple": 1,
        "Banana": 2,
        "Cucumber": 3
    })

    def test_validate(self):
        self.assertEqual(self.validator_1.validate("a"), ValidationResponse.VALID)
        self.assertEqual(self.validator_1.validate("b"), ValidationResponse.VALID)
        self.assertEqual(self.validator_1.validate("c"), ValidationResponse.VALID)
        self.assertEqual(self.validator_1.validate("d"), ValidationResponse.INVALID_ENUM_VALUE)
        self.assertEqual(self.validator_1.validate("abc"), ValidationResponse.INVALID_ENUM_VALUE)
        self.assertEqual(self.validator_1.validate(1.1), ValidationResponse.INVALID_TYPE)
        self.assertEqual(self.validator_2.validate(1), ValidationResponse.VALID)
        self.assertEqual(self.validator_2.validate(2), ValidationResponse.VALID)
        self.assertEqual(self.validator_2.validate(3), ValidationResponse.VALID)
        self.assertEqual(self.validator_2.validate(4), ValidationResponse.INVALID_ENUM_VALUE)
        self.assertEqual(self.validator_2.validate(123), ValidationResponse.INVALID_ENUM_VALUE)

    def test_parse(self):
        val_1 = self.validator_1.parse(self.validator_1.__dict__())

        self.assertEqual(val_1.id, self.validator_1.id)
        self.assertEqual(val_1.name, self.validator_1.name)
        self.assertEqual(val_1.special_id, self.validator_1.special_id)
        self.assertEqual(val_1.values, self.validator_1.values)

        val_2 = self.validator_2.parse(self.validator_2.__dict__())

        self.assertEqual(val_2.id, self.validator_2.id)
        self.assertEqual(val_2.name, self.validator_2.name)
        self.assertEqual(val_2.special_id, self.validator_2.special_id)
        self.assertEqual(val_2.values, self.validator_2.values)

        self.assertEqual(EnumEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": None
        }), EndpointParseResponse.ENUM_TYPE_REQUIRES_VALUES_CONSTRAINT)

        self.assertEqual(EnumEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c"
        }), EndpointParseResponse.ENUM_TYPE_REQUIRES_VALUES_CONSTRAINT)

        self.assertEqual(EnumEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "values": {
                    "a": "Apple",
                    "b": {},
                    "c": "Cucumber"
                }
            }
        }), EndpointParseResponse.CONSTRAINT_VALUES_VALUES_MUST_BE_INT_OR_STRING)

        self.assertEqual(EnumEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "values": {
                    "a": "Apple",
                    "b": "Banana",
                    5: "Cucumber"
                }
            }
        }), EndpointParseResponse.CONSTRAINT_VALUES_KEYS_MUST_BE_STRINGS)

        self.assertEqual(EnumEndpointValidator.parse({
            "id": "a",
            "name": "b",
            "specialId": "c",
            "constraints": {
                "values": 5
            }
        }), EndpointParseResponse.CONSTRAINT_VALUES_MUST_BE_DICT)

    def test__dict__(self):
        data_1 = self.validator_1.__dict__()

        self.assertEqual(data_1.get("id", None), self.validator_1.id)
        self.assertEqual(data_1.get("name", None), self.validator_1.name)
        self.assertEqual(data_1.get("type", None), EndpointType.ENUM.value)
        self.assertEqual(data_1.get("specialId", None), self.validator_1.special_id)
        self.assertIsNotNone(data_1.get("constraints", None))
        self.assertEqual(data_1.get("constraints").get("values", None), self.validator_1.values)

        data_2 = self.validator_2.__dict__()

        self.assertEqual(data_2.get("id", None), self.validator_2.id)
        self.assertEqual(data_2.get("name", None), self.validator_2.name)
        self.assertEqual(data_2.get("type", None), EndpointType.ENUM.value)
        self.assertEqual(data_2.get("specialId", None), self.validator_2.special_id)
        self.assertIsNotNone(data_2.get("constraints", None))
        self.assertEqual(data_2.get("constraints").get("values", None), self.validator_2.values)



# default
import json
import logging
from cgi import parse
from typing import List, Dict, Union, Callable

# external
import flask
from eventlet import websocket
from eventlet.websocket import WebSocket
try:
    from flask_restful import Api

    FLASK_RESTFUL_INSTALLED = True
except ImportError:
    Api = True

    FLASK_RESTFUL_INSTALLED = False

# internal
from .Client import IoTClient
from .Device import DeviceType
from .PacketEncoder import AbstractPacketEncoder, DefaultPacketEncoder
from .types import event_pair, sendable
from .exceptions import ConnectionEnded, ConnectionFailed, ClientNoId, ClientNoType, ClientInvalidType, \
    ClientInvalidData, ClientInvalidEndpoints, ClientNoProtocolVersion, ClientIncompatibleProtocolVersion
from .Errors import Errors
from .Endpoint import EndpointParseResponse, ValidationResponse, AbstractEndpointValidator
from .EndpointResource import EndpointResource
from .__main__ import __protocol_version__


class IoTManagerMiddleware(object):
    def __init__(self, app: flask.Flask, wsgi_app, manager: 'IoTManager'):
        """
        Manager middleware for WSGI.

        :param app: Flask App object.
        :param wsgi_app: Flask App WSGI object.
        :param manager: IoTManager object.
        """
        self.app = app
        self.wsgi_app = wsgi_app
        self.manager = manager

        # eventlet websocket server handler
        self.ws = websocket.WebSocketWSGI(self.socket)

    def __call__(self, environ, start_response):
        # if the path belongs to the iot.io server
        if environ["PATH_INFO"].startswith("/iot.io"):
            # try to handle the request normally
            try:
                with self.app.request_context(environ):
                    # handle using websocket function
                    return self.ws(environ, start_response)
            # if an exception in the handler function is raised deal with it here
            except ConnectionFailed:
                # start response for a 400 code
                start_response("400 Bad Request", [('Content-Type', 'text/plain')])
                return [""]

        # hand down other requests
        return self.wsgi_app(environ, start_response)

    def socket(self, ws):
        self.manager.socket(ws)


# main class used for implementation
class IoTManager(object):
    """
    Main Flask extension.
    """

    def __init__(self, app: flask.Flask, logging_level: int = logging.ERROR, client_logging_level: int = logging.ERROR,
                 encoder: AbstractPacketEncoder = DefaultPacketEncoder, endpoint_api: Union[Api, bool] = None,
                 endpoint_auth_decorator: Callable[[Callable[..., None]], Callable[..., None]] = None):
        """
        A Flask extension used to allow IoT.IO clients to connect to the given flask server.

        :param app: Instance of a Flask object.
        :param logging_level: Logging level of the main IoTManager object.
        :param client_logging_level: Logging level for instances of IoTClient created by the manager.
        :param encoder: Instance of a AbstractPacketEncoder to be used for message serialization/deserialization by
                        instances of IoTClient.
        :param endpoint_api: Either True or an instance of Flask-Restful Api object, used to implement an automatic
                             REST Api for endpoints defined by clients, requires Flask-Restful to be installed.
        :param endpoint_auth_decorator: A wrapper function used to wrap the methods of the Flask-Restful Resource if
                                        user authentication is a desired requirement for accessing the resource.
        """
        # logger for the manager
        self.logger = logging.Logger("iot.io-server")

        # logging levels for main and client debuggers
        self.logger.setLevel(logging_level)
        self.client_logging_level = client_logging_level

        # reference to flask app
        self.app = app

        # reference to Api if it exists
        self.api = None

        # decorator used to authorize
        self.auth_decorator = endpoint_auth_decorator

        # encoder used by clients
        self.encoder = encoder

        # if app is provided initialize the app
        if app:
            self.init_app(app)

            if not hasattr(app, 'extensions'):
                app.extensions = {}  # pragma: no cover
            app.extensions['iot.io'] = self
        else:
            raise ValueError("app must be an instance of a flask application")

        # api object handling
        if endpoint_api is not None:
            if not FLASK_RESTFUL_INSTALLED:
                self.logger.warning("Unable to create Endpoint API, 'flask_restful' must be installed.")

            if isinstance(endpoint_api, Api):
                self.api = endpoint_api
            else:
                if endpoint_api:
                    self.api = Api(self.app, "/api")

        # if the Api exists
        if self.api:
            self.api.add_resource(EndpointResource, "/iot.io", resource_class_kwargs={
                "manager": self,
                "auth_decorator": self.auth_decorator
            })

        # list of active clients
        self.__clients: Dict[str, IoTClient] = {}

        # a dict of device types
        self.__types: Dict[str, DeviceType] = {}

        # dict which is used for storing which clients are in which rooms
        self.__rooms: Dict[str, List[str]] = {}

    # function for adding middleware
    def init_app(self, app: flask.Flask):
        app.wsgi_app = IoTManagerMiddleware(app, app.wsgi_app, self)

    # function for handling new websocket clients
    def socket(self, ws: WebSocket):
        # get headers
        headers = dict(ws.environ["headers_raw"])

        """
        for key, value in headers.items():
            print(key + ": ", value)
        """

        # if id is omitted
        if "IoT-IO-Id" not in headers.keys():
            # send error message to client
            self.__send_error(ws, Errors.CLIENT_NO_ID)

            raise ClientNoId()

        # if type is omitted
        if "IoT-IO-Type" not in headers.keys():
            # send error message to client
            self.__send_error(ws, Errors.CLIENT_NO_TYPE)

            raise ClientNoType()

        # if protocol version is omitted
        if "IoT-IO-ProtocolVersion" not in headers.keys():
            # send error message to client
            self.__send_error(ws, Errors.CLIENT_NO_PROTOCOL_VERSION)

            raise ClientNoProtocolVersion()

        # if data is omitted populate with default value
        if "IoT-IO-Data" not in headers.keys():
            headers["IoT-IO-Data"] = "{}"

        # if endpoints are omitted populate with default value
        if "IoT-IO-Endpoints" not in headers.keys():
            headers["IoT-IO-Endpoints"] = "[]"

        # deserialize headers
        try:
            headers["IoT-IO-Data"] = json.loads(headers["IoT-IO-Data"])
        except json.JSONDecodeError:
            # send error message to client
            self.__send_error(ws, Errors.CLIENT_INVALID_DATA)

            raise ClientInvalidData()

        try:
            headers["IoT-IO-Endpoints"] = json.loads(headers["IoT-IO-Endpoints"])
        except json.JSONDecodeError:
            # send error message to client
            self.__send_error(ws, Errors.CLIENT_INVALID_ENDPOINTS, {
                "endpointId": "",
                "endpointProblem": "invalid_json"
            })

            raise ClientInvalidEndpoints()

        # check if the client's protocol version matches that of the server
        if headers["IoT-IO-ProtocolVersion"] != __protocol_version__:
            # send error message to client
            self.__send_error(ws, Errors.CLIENT_INCOMPATIBLE_PROTOCOL_VERSION)

            raise ClientIncompatibleProtocolVersion()

        # check if the client has a valid type
        if self.__types.get(headers["IoT-IO-Type"], None) is None:
            # log and exit if the client has an invalid type
            self.logger.warning("Client with ID '" + headers["IoT-IO-Id"] + "' claimed a device type of '"
                                + headers["IoT-IO-Type"]
                                + "' but no such device type was defined. Refusing connection.")

            # send error message to client
            self.__send_error(ws, Errors.CLIENT_INVALID_TYPE)

            # abort connection
            raise ClientInvalidType()

        # remove a client with the same id if it exists (getting rid of ghost clients)
        self.remove(headers["IoT-IO-Id"])

        # create a new client
        client = IoTClient(ws, headers["IoT-IO-Id"], headers["IoT-IO-Type"], headers["IoT-IO-Data"], self,
                           logging_level=self.client_logging_level, encoder=self.encoder)

        # parse endpoints
        endpoint_id, response = client.parse_endpoints(headers["IoT-IO-Endpoints"])

        if response != EndpointParseResponse.VALID:
            self.__send_error(ws, Errors.CLIENT_INVALID_ENDPOINTS, {
                "endpointId": endpoint_id,
                "endpointProblem": response.value
            })

            raise ClientInvalidEndpoints()

        # set the request info so we know which client this request belongs to
        flask.request.g = {
            "client": client
        }

        # add the client
        self.add(client)

        # client loop
        try:
            data = ""

            while data is not None:
                # wait for message from the websocket
                data = ws.wait()

                # if the message is None check if the connection is closed
                if data is None:
                    break

                # decode received message
                event, message = self.encoder.decode(data)

                # call the callback associated with the event
                event, response = self.__handle_event_message(event, message, client)

                # check if there is a response to send to the client
                if response is not None:
                    try:
                        client.emit(event, response)
                    except ConnectionEnded:
                        return
        except BrokenPipeError:
            pass
        finally:
            self.remove(client)

    def add(self, client: IoTClient):
        """
        Add a client to the manager.

        :param client: An IoTClient instance.
        :return:
        """
        if not isinstance(client, IoTClient):
            raise TypeError("client must be an instance of IoTClient")

        # add the client to the client list
        self.__clients[client.id] = client

        # call the on_connect handler
        self.__on_connect_handlers(client)

    def remove(self, client: Union[IoTClient, str]):
        """
        Remove a client from the manager.

        :param client: Either the ID of the client as a string or a IoTClient instance
        :return:
        """
        if client is None:
            return False

        if isinstance(client, str):
            if self.__clients.get(client, None) is None:
                return False
            else:
                client = self.__clients[client]
        elif isinstance(client, IoTClient):
            pass
        else:
            raise TypeError("client must be either the id of a client as a str, or an instance of an IoTClient")

        # call the on_disconnect handler
        self.__on_disconnect_handlers(client)

        # remove the client from the rooms
        for room in self.__rooms.keys():
            # remove the client from the room
            self.__rooms[room].remove(client.id)

        # remove the client from the list of clients
        self.__clients.pop(client.id, None)

        return True

    @property
    def clients(self):
        """
        Gets a list of currently connected client IDs.

        :return: A list() of client IDs as strings.
        """
        return self.__clients.keys()

    def get_validator(self, client_id: str, endpoint_id: str) -> Union[AbstractEndpointValidator, ValidationResponse]:
        client = self.__clients.get(client_id, None)

        if client is None:
            return ValidationResponse.CLIENT_NOT_CONNECTED
        else:
            return client.get_validator(endpoint_id)

    def __send_error(self, ws: WebSocket, error: Errors, info: dict = None):
        """
        Send an error to a client via the WebSocket connection.

        :param ws: WebSocket object.
        :param error: Error which has occurred.
        :param info: Additional info about the error if needed.
        :return:
        """
        # log the error being sent
        self.logger.warning("Error being sent to client: '" + str(error.value) + "' | info: '" + str(info) + "'")

        # send the error to the client via the websocket
        ws.send(self.encoder.encode("error", {
            "error": error.value,
            "info": info
        }))

    # handles the device specific and generic on_connect handlers for the specified client
    def __on_connect_handlers(self, client: IoTClient):
        """
        Execute the generic and device specific on_connect handler.

        :param client: New client object.
        :return:
        """

        try:
            # call the type specific handler
            self.__types[client.type].on_connect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling the type specific on_connect handler for client '" + client.id
                              + "': " + str(e))

        try:
            # call the generic handler
            self.on_connect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling generic on_connect handler for client '" + client.id
                              + "': " + str(e))

    # handles the device specific and generic on_disconnect handlers for the specified client
    def __on_disconnect_handlers(self, client: IoTClient):
        """
        Execute the generic and device specific on_disconnect handler.

        :param client: New client object.
        :return:
        """

        try:
            # call the type specific handler
            self.__types[client.type].on_disconnect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling the type specific on_disconnect handler for client '" + client.id
                              + "': " + str(e))

        try:
            # call the generic handler
            self.on_disconnect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling generic on_disconnect handler for client '" + client.id
                              + "': " + str(e))

    # callback for when a client receives a message, sends it to the proper handler and squashes errors
    def __handle_event_message(self, event: str, message: str, client: IoTClient) -> event_pair:
        """
        Handles a given event and message given the client by executing the call_event_handler of the client's type.

        :param event: The event handler to trigger.
        :param message: The message to pass to the handler.
        :param client: The client object to pass to the handler.
        :return:
        """

        # activate the client type's event handler
        return self.__types[client.type].call_event_handler(event, message, client)

    def log_update(self, client: IoTClient, event: str, data: sendable):
        """
        Used by instances of IoTClient to log whenever they send data to

        :param client: Client object representing the client receiving the data.
        :param event: The event for which the data is being sent.
        :param data: The data that was sent to the client.
        :return:
        """
        try:
            self.on_update(client, event, data)
        except Exception as e:
            self.logger.error("Error when calling on_update handler: '" + str(e) + "'")

    def add_type(self, device: DeviceType):
        """
        Define a new device type for the Manager.

        :param device: Any implementation of the subclass DeviceType.
        :return: None
        """

        if not isinstance(device, DeviceType):
            raise ValueError("device is not of type DeviceType")
        elif not isinstance(device.type, str):
            raise ValueError("device.type is not of type str")

        self.__types[device.type] = device
        self.logger.debug("Successfully added DeviceType '" + device.type + "'.")

    def emit(self, event: str, data: sendable, client_id: str = None, client_type: str = None, room: str = None,
             trigger_update: bool = True):
        """
        Emit function for sending data to a single client, group of client types, or room of clients.

        :param event: The client side event to send the data to.
        :param data: The data to be sent to the client(s).
        :param client_id: The id of the client to send the data to, if this param is specified client_type and room are
                          ignored.
        :param client_type: The type of client to send the message to. If it is the only param specified it will send
                            the data to all clients of the specified type, otherwise it acts as a filter on the room
                            specifier.
        :param room: A room id of which all clients should be specified. If client_type is provided it will act as a
                     filter and only send the data ot clients of that type which are also in the specified room.
        :param trigger_update: If this emit should trigger the managers on_update handler. This is useful for preventing
                               unnecessary triggers of the on_update function if for example it is used to update a
                               database for data persistence.
        :return:
        """
        if client_id is not None:
            self.logger.info("emitting event '" + event + "' to client of ID '" + client_id + "'")
            client = self.__clients.get(client_id, None)

            if client:
                client.emit(event, data, trigger_update)
        elif room is not None:
            if self.__rooms.get(room, None) is not None:
                self.logger.info("emitting event '" + event + "' to clients in room '" + room + "'")
                for client_id in self.__rooms[room]:
                    client = self.__clients.get(client_id, None)

                    if client:
                        if client_type is None:
                            client.emit(event, data, trigger_update)
                        else:
                            if client.type == client_type:
                                client.emit(event, data, trigger_update)
        elif client_type is not None:
            self.logger.info("emitting event '" + event + "' to clients of type '" + client_type + "'")
            for client in self.__clients.values():
                if client.type == client_type:
                    client.emit(event, data, trigger_update)
        else:
            raise ValueError("'client_id', 'client_type', or 'room' must be provided")

    def join(self, client: Union[IoTClient, str], room: str, **kwargs):
        """
        Add a client to a specified room.

        :param client: ID of a client as a string or an instance of a IoTClient
        :param room: ID of the room to be joined as a string
        :return:
        """
        if not isinstance(client, str) and not isinstance(client, IoTClient):
            raise TypeError("'client' must be the ID of the client as a str")
        elif not isinstance(room, str):
            raise TypeError("'room' must be an instance of str")

        if isinstance(client, IoTClient):
            client = client.id

        if self.__clients.get(client, None) is None:
            return

        if not kwargs.pop("called_by_client", False):
            self.__clients[client].join(room, called_by_manager=True)

        if self.__rooms.get(room, None) is None:
            self.logger.info("adding client of ID '" + client + "' to newly created room '" + room + "'")
            self.__rooms[room] = [client]
        else:
            self.logger.info("adding client of ID '" + client + "' to existing room '" + room + "'")
            self.__rooms[room].append(client)

    def leave(self, client: Union[IoTClient, str], room: str, **kwargs):
        """
        Remove a client from a specified room.

        :param client: ID of a client as a string or an instance of a IoTClient
        :param room: ID of the room to be left as a string
        :return:
        """
        if not isinstance(client, str):
            raise TypeError("'client' must be the ID of the client as a str")
        elif not isinstance(room, str):
            raise TypeError("'room' must be an instance of str")

        if isinstance(client, IoTClient):
            client = client.id

        if self.__clients.get(client, None) is None:
            return

        if not kwargs.pop("called_by_client", False):
            self.__clients[client].leave(room, called_by_manager=True)

        if self.__rooms.get(room, None) is None:
            return
        else:
            self.logger.info("removing client of ID '" + client + "' from room '" + room + "'")
            self.__rooms[room].remove(client)

    def close_room(self, room: str):
        """
        Close a room and remove all of it's clients.

        :param room: ID of the room to be closed
        :return:
        """
        if not isinstance(room, str):
            raise TypeError("'room' must be an instance of str")

        rooms = self.__rooms.pop(room, None)

        if rooms:
            self.logger.info("closing '" + room + "'")
            for client in rooms:
                client = self.__clients.get(client, None)
                if client:
                    client.leave(room, called_by_manager=True)
        else:
            self.logger.info("attempted to close room '" + room + "' but it did not exist")

    # event decorator function
    def event(self, coroutine):
        """
        Decorator which allows for simple overwriting of the generic on_connect, and
        on_disconnect functions.

        :param coroutine: Function named on_connect, on_disconnect, or on_update.
        :return: bool
        """

        # handle general on_connect, and on_disconnect, and on_update handlers
        if coroutine.__name__ == "on_connect" or coroutine.__name__ == "on_disconnect" or \
                coroutine.__name__ == "on_update":
            # logging output
            self.logger.info("Event handler '" + coroutine.__name__ + "' was added successfully.")

            # replaces the existing coroutine with the provided one
            setattr(self, coroutine.__name__, coroutine)
            return True
        return False

    def on_connect(self, client: IoTClient):
        """
        An empty implementation of the generic on_connect function, meant to be overwritten.

        Runs when any client connects to the server.

        :param client: Client object representing the connecting client.
        :return: None or str.
        """
        pass

    def on_disconnect(self, client: IoTClient):
        """
        An empty implementation of the generic on_disconnect function, meant to be overwritten.

        Runs when any client disconnects from the server.

        :param client: Client object representing the disconnecting client.
        :return: None.
        """
        pass

    def on_update(self, client: IoTClient, event: str, data: sendable):
        """
        An empty implementation of the on_update function, meant to be written overwritten.

        Runs whenever data is emitted from the server to a client.

        :param client: Client object representing the client receiving the data.
        :param event: The event for which the data is being sent.
        :param data: The data that was sent to the client.
        :return:
        """
        pass


def emit(event: str, data: sendable, client_id: str = None, client_type: str = None, room: str = None,
         trigger_update: bool = True):
    """
    Emit function for sending data to a single client, group of client types, or room of clients.

    :param event: The client side event to send the data to.
    :param data: The data to be sent to the client(s).
    :param client_id: The id of the client to send the data to, if this param is specified client_type and room are
                      ignored.
    :param client_type: The type of client to send the message to. If it is the only param specified it will send
                        the data to all clients of the specified type, otherwise it acts as a filter on the room
                        specifier.
    :param room: A room id of which all clients should be specified. If client_type is provided it will act as a
                 filter and only send the data ot clients of that type which are also in the specified room.
    :param trigger_update: If this emit should trigger the managers on_update handler. This is useful for preventing
                           unnecessary triggers of the on_update function if for example it is used to update a database
                           for data persistence.
    :return:
    """
    if client_id is None and client_type is None and room is None:
        client_id = flask.request.g["client"].id

    return flask.current_app.extensions["iot.io"].emit(event, data, client_id, client_type, room, trigger_update)


def join(room: str, client: Union[IoTClient, str] = None):
    """
    Add a client to a specified room.

    :param client: ID of a client as a string or an instance of a IoTClient
    :param room: ID of the room to be joined as a string
    :return:
    """
    if client is None:
        client = flask.request.g["client"].id

    return flask.current_app.extensions["iot.io"].join(client, room)


def leave(room: str, client: Union[IoTClient, str] = None):
    """
    Remove a client from a specified room.

    :param client: ID of a client as a string or an instance of a IoTClient
    :param room: ID of the room to be left as a string
    :return:
    """
    if client is None:
        client = flask.request.g["client"].id

    return flask.current_app.extensions["iot.io"].leave(client, room)


def close_room(room: str):
    """
    Close a room and remove all of it's clients.

    :param room: ID of the room to be closed
    :return:
    """
    return flask.current_app.extensions["iot.io"].close_room(room)



# default
from typing import Union, Dict, List, Tuple
import enum

filter_list = Union[str, List[str]]
enum_list = Dict[str, str]


class ValidationResponse(enum.Enum):
    VALID = "valid"
    INVALID_TYPE = "invalid_data_type"
    LOWER_THAN_MIN = "lower_than_min"
    HIGHER_THAN_MAX = "higher_than_max"
    INVALID_INCREMENT = "invalid_increment"
    INVALID_LENGTH = "invalid_length"
    VIOLATES_ALLOW_LIST = "violates_allow_list"
    VIOLATES_DENY_LIST = "violates_deny_list"
    INVALID_ENUM_VALUE = "invalid_enum_value"
    ENDPOINT_NOT_DEFINED_BY_CLIENT = "endpoint_not_defined_by_client"
    CLIENT_ENDPOINTS_NOT_INITIALIZED = "client_endpoints_not_initialized"
    CLIENT_NOT_CONNECTED = "client_not_connected"


class EndpointParseResponse(enum.Enum):
    VALID = "valid"
    NO_ID_SPECIFIED = "no_id_specified"
    NO_TYPE_SPECIFIED = "no_type_specified"
    NO_NAME_SPECIFIED = "no_name_specified"
    INVALID_TYPE_SPECIFIED = "invalid_type_specified"
    ID_MUST_BE_STRING = "id_must_be_string"
    TYPE_MUST_BE_STRING = "type_must_be_string"
    NAME_MUST_BE_STRING = "name_must_be_string"
    SPECIAL_ID_MUST_BE_STRING = "special_id_must_be_string"
    CONSTRAINT_MIN_MUST_BE_INT = "constraint_min_must_be_integer"
    CONSTRAINT_MAX_MUST_BE_INT = "constraint_max_must_be_integer"
    CONSTRAINT_INCREMENT_MUST_BE_INT = "constraint_increment_must_be_integer"
    CONSTRAINT_LENGTH_MUST_BE_INT = "constraint_length_must_be_integer"
    CONSTRAINT_BOTH_ALLOW_LIST_AND_DENY_LIST_SPECIFIED = "constraint_both_allow_list_and_deny_list_specified"
    CONSTRAINT_LIST_MUST_BE_STRING_OR_LIST = "constraint_list_must_be_string_or_list"
    CONSTRAINT_LIST_MUST_ONLY_CONTAIN_STRINGS = "constraint_list_must_only_contain_strings"
    ENUM_TYPE_REQUIRES_VALUES_CONSTRAINT = "enum_type_requires_values_constraint"
    CONSTRAINT_VALUES_MUST_BE_DICT = "constraint_values_must_be_dict"
    CONSTRAINT_VALUES_KEYS_MUST_BE_STRINGS = "constraint_values_keys_must_be_strings"
    CONSTRAINT_VALUES_VALUES_MUST_BE_INT_OR_STRING = "constraint_values_values_must_be_int_or_string"


class EndpointType(enum.Enum):
    BOOLEAN = "boolean"
    INTEGER = "integer"
    STRING = "string"
    ENUM = "enum"


# abstract validator class
class AbstractEndpointValidator:
    def __init__(self, endpoint_id: str, name: str, special_id: Union[str, None]):
        self.id = endpoint_id
        self.name = name
        self.special_id = special_id
        self.type = None

    def validate(self, data) -> ValidationResponse:
        pass

    def __iter__(self):
        yield "id", self.id
        yield "name", self.name
        yield "specialId", self.special_id
        yield "constraints", None

    @staticmethod
    def parse(endpoint_data: dict) -> Union['AbstractEndpointValidator', EndpointParseResponse]:
        if "id" not in endpoint_data:
            return EndpointParseResponse.NO_ID_SPECIFIED
        elif "name" not in endpoint_data:
            return EndpointParseResponse.NO_NAME_SPECIFIED

        # ensure valid types for id, name, and specialId
        if not isinstance(endpoint_data.get("id", ""), str):
            return EndpointParseResponse.ID_MUST_BE_STRING

        if not isinstance(endpoint_data.get("name", ""), str):
            return EndpointParseResponse.NAME_MUST_BE_STRING

        if not isinstance(endpoint_data.get("specialId", ""), str) and not endpoint_data.get("specialId", None) is None:
            return EndpointParseResponse.SPECIAL_ID_MUST_BE_STRING

        return AbstractEndpointValidator(
            endpoint_data["id"],
            endpoint_data["name"],
            endpoint_data.get("specialId", None)
        )


class BooleanEndpointValidator(AbstractEndpointValidator):
    def __init__(self, endpoint_id: str, name: str, special_id: Union[str, None]):
        super().__init__(endpoint_id, name, special_id)
        self.type = EndpointType.BOOLEAN

    def validate(self, data) -> Tuple[ValidationResponse, Union[bool, None]]:
        # check that data is a boolean value
        if not isinstance(data, bool):
            try:
                if isinstance(data, str):
                    if data.lower() == "true":
                        data = True
                    elif data.lower() == "false":
                        data = False
                    else:
                        return ValidationResponse.INVALID_TYPE, None
                else:
                    data = bool(data)
            except ValueError:
                return ValidationResponse.INVALID_TYPE, None

        return ValidationResponse.VALID, data

    def __iter__(self):
        yield "id", self.id
        yield "name", self.name
        yield "type", self.type.value
        yield "specialId", self.special_id
        yield "constraints", None

    @staticmethod
    def parse(endpoint_data: dict) -> Union['BooleanEndpointValidator', EndpointParseResponse]:
        validator = AbstractEndpointValidator.parse(endpoint_data)

        # if the parse failed then return the EndpointParseResponse
        if not isinstance(validator, AbstractEndpointValidator):
            return validator

        # return the validator
        return BooleanEndpointValidator(
            validator.id,
            validator.name,
            validator.special_id
        )


class IntegerEndpointValidator(AbstractEndpointValidator):
    def __init__(self, endpoint_id: str, name: str, special_id: Union[str, None],
                 min_v: Union[None, int] = None, max_v: Union[None, int] = None,
                 increment: Union[None, int] = None):
        super().__init__(endpoint_id, name, special_id)
        self.type = EndpointType.INTEGER
        self.min = min_v
        self.max = max_v
        self.increment = increment

    def validate(self, data) -> Tuple[ValidationResponse, Union[int, None]]:
        # check if it is a number
        if not isinstance(data, int):
            try:
                data = int(data)
            except ValueError:
                return ValidationResponse.INVALID_TYPE, None

        # check if a minimum value was provided
        if self.min is not None:
            # check if it is less than the minimum
            if data < self.min:
                return ValidationResponse.LOWER_THAN_MIN, None

        # check if a maximum value was provided
        if self.max is not None:
            # check if it is more than the maximum
            if data > self.max:
                return ValidationResponse.HIGHER_THAN_MAX, None

        return ValidationResponse.VALID, data

    def __iter__(self):
        yield "id", self.id
        yield "name", self.name
        yield "type", self.type.value
        yield "specialId", self.special_id
        yield "constraints", {
            "min": self.min,
            "max": self.max,
            "increment": self.increment
        }

    @staticmethod
    def parse(endpoint_data: dict) -> Union['IntegerEndpointValidator', EndpointParseResponse]:
        validator = AbstractEndpointValidator.parse(endpoint_data)

        # if the parse failed then return the EndpointParseResponse
        if not isinstance(validator, AbstractEndpointValidator):
            return validator

        # check that constraints were provided
        if endpoint_data.get("constraints", None) is None:
            constraints = {}
        else:
            constraints = endpoint_data.get("constraints", {})

            # check that the min, max, and increment constraints are all integers or None
            if constraints.get("min", None) is not None:
                if not isinstance(constraints.get("min", 0), int):
                    return EndpointParseResponse.CONSTRAINT_MIN_MUST_BE_INT

            if constraints.get("max", None) is not None:
                if not isinstance(constraints.get("max", 0), int):
                    return EndpointParseResponse.CONSTRAINT_MAX_MUST_BE_INT

            if constraints.get("increment", None) is not None:
                if not isinstance(constraints.get("increment", 0), int):
                    return EndpointParseResponse.CONSTRAINT_INCREMENT_MUST_BE_INT

        # return the validator
        return IntegerEndpointValidator(
            endpoint_id=validator.id,
            name=validator.name,
            special_id=validator.special_id,
            min_v=constraints.get("min", None),
            max_v=constraints.get("max", None),
            increment=constraints.get("increment", None)
        )


class StringEndpointValidator(AbstractEndpointValidator):
    def __init__(self, endpoint_id: str, name: str, special_id: Union[str, None],
                 length: Union[None, int] = None, allow_list: Union[None, filter_list] = None,
                 deny_list: Union[None, filter_list] = None):
        super().__init__(endpoint_id, name, special_id)
        self.type = EndpointType.STRING
        self.length = length
        self.allow_list = allow_list
        self.deny_list = deny_list

    def validate(self, data) -> Tuple[ValidationResponse, Union[str, None]]:
        # check if it is a string
        if not isinstance(data, str):
            try:
                data = str(data)
            except ValueError:
                return ValidationResponse.INVALID_TYPE, None

        # check if the length constraint was provided
        if self.length is not None:
            # check if the length is correct
            if len(data) != self.length:
                return ValidationResponse.INVALID_LENGTH, None

        # check if allow_list is provided
        if self.allow_list is not None:
            if isinstance(self.allow_list, str):
                # for each char in the provided string
                for char in data:
                    # check if it is not in the allow_list
                    if char not in self.allow_list:
                        return ValidationResponse.VIOLATES_ALLOW_LIST, None
            # if the allow_list is an array of strings
            else:
                # if allow_list is a list of strings check if provided string matches one of those strings
                if data not in self.allow_list:
                    return ValidationResponse.VIOLATES_ALLOW_LIST, None

        # check if deny_list is provided
        if self.deny_list is not None:
            if isinstance(self.deny_list, str):
                # for each char in the deny list
                for char in self.deny_list:
                    # check if the deny_list contains that char
                    if char in data:
                        return ValidationResponse.VIOLATES_DENY_LIST, None
            else:
                if data in self.deny_list:
                    return ValidationResponse.VIOLATES_DENY_LIST, None

        return ValidationResponse.VALID, data

    def __iter__(self):
        yield "id", self.id
        yield "name", self.name
        yield "type", self.type.value
        yield "specialId", self.special_id
        yield "constraints", {
            "length": self.length,
            "allowList": self.allow_list,
            "denyList": self.deny_list
        }

    @staticmethod
    def parse(endpoint_data: dict) -> Union['StringEndpointValidator', EndpointParseResponse]:
        # define check list func
        def check_list(data: dict, list_name: str) -> EndpointParseResponse:
            if data.get(list_name, "") is None:
                return EndpointParseResponse.VALID

            if isinstance(data.get(list_name, ""), str) or isinstance(data.get(list_name, []), list):
                if isinstance(data.get(list_name, None), list):
                    for item in data.get(list_name, []):
                        if not isinstance(item, str):
                            return EndpointParseResponse.CONSTRAINT_LIST_MUST_ONLY_CONTAIN_STRINGS
            else:
                return EndpointParseResponse.CONSTRAINT_LIST_MUST_BE_STRING_OR_LIST

            return EndpointParseResponse.VALID

        validator = AbstractEndpointValidator.parse(endpoint_data)

        # if the parse failed then return the EndpointParseResponse
        if not isinstance(validator, AbstractEndpointValidator):
            return validator

        # check that constraints were provided
        if endpoint_data.get("constraints", None) is None:
            constraints = {}
        else:
            constraints = endpoint_data.get("constraints", {})

            # check that the length field is an int or none
            if constraints.get("length", None) is not None:
                if not isinstance(constraints.get("length", 0), int):
                    return EndpointParseResponse.CONSTRAINT_LENGTH_MUST_BE_INT

            # ensure that allow_list and deny_list are not provided at the same time
            if constraints.get("allowList", None) is not None and constraints.get("denyList", None) is not None:
                return EndpointParseResponse.CONSTRAINT_BOTH_ALLOW_LIST_AND_DENY_LIST_SPECIFIED

            # check if allow_list and deny_list are formatted correctly
            result = check_list(constraints, "allowList")

            if result is not EndpointParseResponse.VALID:
                return result

            result = check_list(constraints, "denyList")

            if result is not EndpointParseResponse.VALID:
                return result

        # return the validator
        return StringEndpointValidator(
            endpoint_id=validator.id,
            name=validator.name,
            special_id=validator.special_id,
            length=constraints.get("length", None),
            allow_list=constraints.get("allowList", None),
            deny_list=constraints.get("denyList", None)
        )


class EnumEndpointValidator(AbstractEndpointValidator):
    def __init__(self, endpoint_id: str, name: str, special_id: Union[str, None],
                 values: Dict[str, Union[int, str]]):
        super().__init__(endpoint_id, name, special_id)
        self.type = EndpointType.ENUM
        self.values = values

    def validate(self, data) -> Tuple[ValidationResponse, Union[str, int, None]]:
        # check if it is a string
        if not isinstance(data, str) and not isinstance(data, int):
            try:
                data = str(data)
            except ValueError:
                return ValidationResponse.INVALID_TYPE, None

        # check if the value is a valid enum for this endpoint
        if data not in self.values.values():
            try:
                data = int(data)

                if data not in self.values.values():
                    return ValidationResponse.INVALID_ENUM_VALUE, None
            except ValueError:
                return ValidationResponse.INVALID_ENUM_VALUE, None

        return ValidationResponse.VALID, data

    def __iter__(self):
        yield "id", self.id
        yield "name", self.name
        yield "type", self.type.value
        yield "specialId", self.special_id
        yield "constraints", {
            "values": self.values
        }

    @staticmethod
    def parse(endpoint_data: dict) -> Union['EnumEndpointValidator', EndpointParseResponse]:
        validator = AbstractEndpointValidator.parse(endpoint_data)

        # if the parse failed then return the EndpointParseResponse
        if not isinstance(validator, AbstractEndpointValidator):
            return validator

        # check that constraints were provided
        if endpoint_data.get("constraints", None) is None:
            return EndpointParseResponse.ENUM_TYPE_REQUIRES_VALUES_CONSTRAINT
        else:
            constraints = endpoint_data.get("constraints")
            constraints: dict

            # ensure the "values" value exists
            if constraints.get("values", None) is None:
                return EndpointParseResponse.ENUM_TYPE_REQUIRES_VALUES_CONSTRAINT
            else:
                # ensure it is a dictionary
                if not isinstance(constraints.get("values"), dict):
                    return EndpointParseResponse.CONSTRAINT_VALUES_MUST_BE_DICT

            # ensure that both the key and value for each entry in the "values" key are strings
            for key, val in constraints.get("values").items():
                if not isinstance(key, str):
                    return EndpointParseResponse.CONSTRAINT_VALUES_KEYS_MUST_BE_STRINGS
                elif not isinstance(val, str) and not isinstance(val, int):
                    return EndpointParseResponse.CONSTRAINT_VALUES_VALUES_MUST_BE_INT_OR_STRING

        # return the validator
        return EnumEndpointValidator(
            endpoint_id=validator.id,
            name=validator.name,
            special_id=validator.special_id,
            values=constraints.get("values")
        )


class EndpointManager:
    def __init__(self):
        self.__endpoints = {}
        self.__initialized = False

    @property
    def initialized(self):
        return self.__initialized

    @property
    def endpoints(self):
        return [dict(x) for x in self.__endpoints.values()]

    def __add_endpoint(self, validator: AbstractEndpointValidator):
        self.__endpoints[validator.id] = validator

    def get_validator(self, endpoint_id: str) -> Union[AbstractEndpointValidator, ValidationResponse]:
        self.__endpoints: Dict[str, AbstractEndpointValidator]

        try:
            return self.__endpoints[endpoint_id]
        except KeyError:
            return ValidationResponse.ENDPOINT_NOT_DEFINED_BY_CLIENT

    def parse(self, endpoints: List[dict]) -> Tuple[str, EndpointParseResponse]:
        for endpoint in endpoints:
            # parse the endpoint
            if endpoint.get("type", None) is None:
                return endpoint.get("id", "no_id_provided"), EndpointParseResponse.NO_TYPE_SPECIFIED

            # get validator based on type
            if endpoint["type"] == EndpointType.BOOLEAN.value:
                validator = BooleanEndpointValidator.parse(endpoint)
            elif endpoint["type"] == EndpointType.INTEGER.value:
                validator = IntegerEndpointValidator.parse(endpoint)
            elif endpoint["type"] == EndpointType.STRING.value:
                validator = StringEndpointValidator.parse(endpoint)
            elif endpoint["type"] == EndpointType.ENUM.value:
                validator = EnumEndpointValidator.parse(endpoint)
            else:
                # if an invalid type was specified
                return endpoint.get("id", "no_id_provided"), EndpointParseResponse.INVALID_TYPE_SPECIFIED

            # if the parse failed then return the EndpointParseResponse
            if not isinstance(validator, AbstractEndpointValidator):
                return endpoint.get("id", "no_id_provided"), validator

            # if parse was successful then add the validator to the manager
            self.__add_endpoint(validator)

        # set initialized flag
        self.__initialized = True

        return "", EndpointParseResponse.VALID



from unittest import TestCase
from flask import Flask
from iotio import IoTManager, IoTClient
from iotio.PacketEncoder import DefaultPacketEncoder
from iotio.types import sendable
from eventlet.websocket import WebSocket
from typing import List


# used for testing TestIotClient, pretends to be a functioning IoTManager
class TestIoTManager(IoTManager):
    def __init__(self):
        app = Flask("")
        super().__init__(app)

    def log_update(self, client: IoTClient, event: str, data: sendable):
        pass


# used for testing TestIoTClient, pretends to be a functioning eventlet WebSocket object
class TestWebSocket(WebSocket):
    def __init__(self, messages: List[str] = None):
        self.messages = []
        self.closed = False

        if messages is None:
            messages = ["abc", "def", "ghi"]

        messages.append("stop")

        for m in messages:
            self.messages.append(DefaultPacketEncoder.encode("test", m))

        super().__init__(None, {})

    @property
    def websocket_closed(self):
        return self.closed

    @websocket_closed.setter
    def websocket_closed(self, data: bool):
        self.closed = data

    def send(self, data: bytearray):
        pass

    def wait(self):
        return self.messages.pop(0)


class TestIoTClient(TestCase):
    c_socket = TestWebSocket()
    c_id = "test_client"
    c_type = "test"
    c_data = {
        "count": 5
    }
    c_endpoints = []

    client = IoTClient(c_socket, c_id, c_type, c_data, TestIoTManager())

    def test_id(self):
        self.assertEqual(self.client.id, self.c_id)

    def test_type(self):
        self.assertEqual(self.client.type, self.c_type)

    def test_data(self):
        self.assertEqual(self.client.data, self.c_data)

    # noinspection PyBroadException
    def test_send(self):
        try:
            self.client.emit("test", "data")
        except Exception:
            self.fail()

        self.c_socket.websocket_closed = True
        try:
            self.client.emit("test", "data")
            self.fail()
        except Exception:
            pass
        self.c_socket.websocket_closed = False



from setuptools import setup
import iotio

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="iot.io",
    version=iotio.__version__,
    packages=["iotio"],
    author="Dylan Crockett",
    author_email="dylanrcrockett@gmail.com",
    license="MIT",
    description="A management API for connecting and managing Clients via websocket connections.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dylancrockett/iot.io",
    project_urls={
        "Documentation": "https://iotio.readthedocs.io/",
        "Source Code": "https://github.com/dylancrockett/iot.io"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'gevent',
        'gevent-websocket',
        'flask',
        'flask-sockets',
    ],
    python_requires='>=3.7'
)



from .Manager import IoTManager, emit, join, leave, close_room
from .Client import IoTClient
from .Device import DeviceType

__title = "iot.io"
__author__ = "Dylan Crockett"
__licence__ = "MIT"
__version__ = "0.4.7"
